from datetime import datetime
from os.path import isfile
from typing import Any, Optional, Type, TypedDict, TypeVar, Union

import ruyaml as yaml

from pyhilo.const import LOG


class TokenDict(TypedDict):
    access: Optional[str]
    refresh: Optional[str]
    expires_at: datetime


class AndroidDeviceDict(TypedDict):
    token: str
    device_id: int


class WebsocketTransportsDict(TypedDict):
    transport: str
    transfer_formats: list[str]


class WebsocketDict(TypedDict, total=False):
    token: str
    connection_id: str
    full_ws_url: str
    url: str
    available_transports: list[WebsocketTransportsDict]


class RegistrationDict(TypedDict, total=False):
    reg_id: str
    expires_at: datetime


class FirebaseDict(TypedDict):
    fid: Optional[str]
    name: Optional[str]  # "projects/18450192328/installations/d7N8yHopRWOiTYCrnYLi8a"
    token: TokenDict


class StateDict(TypedDict, total=False):
    token: TokenDict
    registration: RegistrationDict
    firebase: FirebaseDict
    android: AndroidDeviceDict
    websocket: WebsocketDict


T = TypeVar("T", bound="StateDict")


def __get_defaults__(cls: Type[T]) -> dict[str, Any]:
    """Generates a default dict based on typed dict

    :param cls: TypedDict class
    :type cls: Type[T]
    :return: Dictionary with empty values
    :rtype: dict[str, Any]
    """
    # NOTE(dvd): Find a better way of identifying another TypedDict.
    new_dict: StateDict = {}
    for k, v in cls.__annotations__.items():
        if hasattr(v, "__annotations__"):
            new_dict[k] = __get_defaults__(v)  # type: ignore
        else:
            new_dict[k] = None  # type: ignore
    return new_dict  # type: ignore


def get_state(state_yaml: str) -> StateDict:
    """Read in state yaml.
    :param state_yaml: filename where to read the state
    :type state_yaml: ``str``
    :rtype: ``StateDict``
    """
    if not isfile(state_yaml):
        return __get_defaults__(StateDict)  # type: ignore
    with open(state_yaml) as yaml_file:
        LOG.debug("Loading state from yaml")
        state_yaml_payload: StateDict = yaml.load(yaml_file, Loader=yaml.Loader)
    return state_yaml_payload


def set_state(
    state_yaml: str,
    key: str,
    state: Union[
        TokenDict, RegistrationDict, FirebaseDict, AndroidDeviceDict, WebsocketDict
    ],
) -> None:
    """Save state yaml.
    :param state_yaml: filename where to read the state
    :type state_yaml: ``str``
    :param key: Key name
    :type key: ``str``
    :param state: Dictionary containing the state
    :type state: ``StateDict``
    :rtype: ``StateDict``
    """
    current_state = get_state(state_yaml)
    merged_state: dict[str, Any] = {key: {**current_state.get(key, {}), **state}}  # type: ignore
    new_state: dict[str, Any] = {**current_state, **merged_state}
    with open(state_yaml, "w") as yaml_file:
        LOG.debug("Saving state to yaml file")
        yaml.dump(new_state, yaml_file, Dumper=yaml.RoundTripDumper)



"""Event object """
from datetime import datetime, timedelta
import re
from typing import Any, cast

from pyhilo.const import LOG
from pyhilo.util import camel_to_snake, from_utc_timestamp


class Event:
    preheat_start: datetime
    preheat_end: datetime
    reduction_start: datetime
    reduction_end: datetime
    recovery_start: datetime
    recovery_end: datetime

    def __init__(self, **event: dict[str, Any]):
        self._convert_phases(cast(dict[str, Any], event.get("phases")))
        params: dict[str, Any] = event.get("parameters", {})
        devices: list[dict[str, Any]] = params.get("devices", [])
        consumption: dict[str, Any] = event.get("consumption", {})
        allowed_wH: int = consumption.get("baselineWh", 0) or 0
        used_wH: int = consumption.get("currentWh", 0) or 0
        self.participating: bool = cast(bool, event.get("isParticipating", False))
        self.configurable: bool = cast(bool, event.get("isConfigurable", False))
        self.period: str = cast(str, event.get("period", ""))
        self.event_id: int = cast(int, event["id"])
        self.total_devices: int = len(devices)
        self.opt_out_devices: int = len([x for x in devices if x["optOut"]])
        self.pre_heat_devices: int = len([x for x in devices if x["preheat"]])
        self.progress: str = cast(str, event.get("progress", "unknown"))
        self.mode: str = cast(str, params.get("mode", "Unknown"))
        self.allowed_kWh: float = round(allowed_wH / 1000, 2)
        self.used_kWh: float = round(used_wH / 1000, 2)
        self.used_percentage: float = 0
        if allowed_wH > 0:
            self.used_percentage = round(used_wH / allowed_wH * 100, 2)
        self.dict_items = [
            "participating",
            "configurable",
            "period",
            "total_devices",
            "opt_out_devices",
            "pre_heat_devices",
            "mode",
            "allowed_kWh",
            "used_kWh",
            "used_percentage",
        ]
        LOG.debug(f"Adding event for {self.phases_list[0]}")

    def as_dict(self) -> dict[str, Any]:
        rep = {k: getattr(self, k) for k in self.dict_items}
        rep["phases"] = {k: getattr(self, k) for k in self.phases_list}
        rep["state"] = self.state
        return rep

    def _convert_phases(self, phases: dict[str, Any]) -> None:
        if not len(phases):
            return
        self.phases_list = []
        for key, value in phases.items():
            if not key.endswith("DateUTC"):
                continue
            phase_match = re.match(r"(.*)DateUTC", key)
            if not phase_match:
                continue
            phase = camel_to_snake(phase_match.group(1))
            setattr(self, phase, from_utc_timestamp(value))
            self.phases_list.append(phase)

    def appreciation(self, hours: int) -> datetime:
        """Wrapper to return X hours before pre_heat.
        Will also set appreciation_start and appreciation end phases.
        """
        LOG.debug(f"Adding appreciation phases: {hours}")
        self.appreciation_start = self.preheat_start - timedelta(hours=hours)
        self.appreciation_end = self.preheat_start
        if "appreciation_start" not in self.phases_list:
            self.phases_list[:0] = ["appreciation_start", "appreciation_end"]
        return self.appreciation_start

    @property
    def state(self) -> str:
        now = datetime.now(self.preheat_start.tzinfo)
        if self.preheat_start > now:
            return "scheduled"
        elif self.preheat_start <= now < self.preheat_end:
            return "pre_heat"
        elif self.reduction_start <= now < self.reduction_end:
            return "reduction"
        elif self.recovery_start <= now < self.recovery_end:
            return "recovery"
        elif now <= self.recovery_end:
            return "completed"
        elif self.progress:
            return self.progress
        else:
            return "unknown"



"""Define package errors."""
from __future__ import annotations


class HiloError(Exception):
    """A base error."""

    pass


class EndpointUnavailableError(HiloError):
    """An error related to accessing an endpoint that isn't available in the plan."""

    pass


class InvalidCredentialsError(HiloError):
    """An error related to invalid credentials."""

    pass


class RequestError(HiloError):
    """An error related to invalid requests."""

    pass


class WebsocketError(HiloError):
    """An error related to generic websocket errors."""

    pass


class CannotConnectError(WebsocketError):
    """Define a error when the websocket can't be connected to."""

    pass


class ConnectionClosedError(WebsocketError):
    """Define a error when the websocket closes unexpectedly."""

    pass


class ConnectionFailedError(WebsocketError):
    """Define a error when the websocket connection fails."""

    pass


class InvalidMessageError(WebsocketError):
    """Define a error related to an invalid message from the websocket server."""

    pass


class NotConnectedError(WebsocketError):
    """Define a error when the websocket isn't properly connected to."""

    pass



"""Define a connection to the Hilo websocket."""
from __future__ import annotations

import asyncio
from dataclasses import dataclass, field
from datetime import datetime, timedelta
from enum import IntEnum
import json
from os import environ
from typing import TYPE_CHECKING, Any, Callable, Dict, cast

from aiohttp import ClientWebSocketResponse, WSMsgType
from aiohttp.client_exceptions import (
    ClientError,
    ServerDisconnectedError,
    WSServerHandshakeError,
)
from yarl import URL

from pyhilo.const import DEFAULT_USER_AGENT, LOG
from pyhilo.exceptions import (
    CannotConnectError,
    ConnectionClosedError,
    ConnectionFailedError,
    InvalidCredentialsError,
    InvalidMessageError,
    NotConnectedError,
)
from pyhilo.util import schedule_callback

if TYPE_CHECKING:
    from pyhilo import API

DEFAULT_WATCHDOG_TIMEOUT = timedelta(minutes=5)


class SignalRMsgType(IntEnum):
    INVOKE = 0x1
    STREAM = 0x2
    COMPLETE = 0x3
    STREAM_INVOCATION = 0x4
    CANCEL_INVOCATION = 0x5
    PING = 0x6
    CLOSE = 0x7
    UNKNOWN = 0xFF

    @classmethod
    def has_value(cls, value: int) -> bool:
        return value in cls._value2member_map_

    @classmethod
    def value(cls, value: int) -> IntEnum:  # type: ignore
        return cls._value2member_map_.get(value, cls.UNKNOWN)  # type: ignore


@dataclass(frozen=True)
class WebsocketEvent:
    """Define a representation of a message."""

    event_type_id: int
    target: str
    arguments: list[list]
    invocation: int | None
    error: str | None
    timestamp: datetime = field(default=datetime.now())
    event_type: str | None = field(init=False)

    def __post_init__(self) -> None:
        if SignalRMsgType.has_value(self.event_type_id):
            object.__setattr__(
                self, "event_type", SignalRMsgType.value(self.event_type_id).name
            )
        if self.event_type_id == SignalRMsgType.CLOSE:
            LOG.error(
                f"Received close event from SignalR: Error: {self.event_type} Target: {self.target} Args: {self.arguments} Error: {self.error}"
            )


def websocket_event_from_payload(payload: dict[str, Any]) -> WebsocketEvent:
    """Create a Message object from a websocket event payload."""
    return WebsocketEvent(
        payload["type"],
        payload.get("target", ""),
        payload.get("arguments", ""),
        payload.get("invocationId"),
        payload.get("error"),
    )


class Watchdog:
    """Define a watchdog to kick the websocket connection at intervals."""

    def __init__(
        self, action: Callable[..., Any], timeout: timedelta = DEFAULT_WATCHDOG_TIMEOUT
    ):
        """Initialize."""
        self._action = action
        self._action_task: asyncio.Task | None = None
        self._loop = asyncio.get_running_loop()
        self._timeout_seconds = timeout.total_seconds()
        self._timer_task: asyncio.TimerHandle | None = None

    def _on_expire(self) -> None:
        """Log and act when the watchdog expires."""
        LOG.warning("Websocket: Watchdog expired")
        schedule_callback(self._action)

    def cancel(self) -> None:
        """Cancel the watchdog."""
        if self._timer_task:
            self._timer_task.cancel()
            self._timer_task = None

    def trigger(self) -> None:
        """Trigger the watchdog."""
        if self._timer_task:
            self._timer_task.cancel()

        self._timer_task = self._loop.call_later(self._timeout_seconds, self._on_expire)


class WebsocketClient:
    """A websocket connection to the Hilo cloud.
    Note that this class shouldn't be instantiated directly; it will be instantiated as
    :param api: A :meth:`pyhilo.API` object
    :type api: :meth:`pyhilo.API`
    """

    def __init__(self, api: API) -> None:
        """Initialize."""
        self._api = api
        self._connect_callbacks: list[Callable[..., None]] = []
        self._disconnect_callbacks: list[Callable[..., None]] = []
        self._event_callbacks: list[Callable[..., None]] = []
        self._loop = asyncio.get_running_loop()
        self._watchdog = Watchdog(self.async_reconnect)
        self._ready_event: asyncio.Event = asyncio.Event()
        self._ready: bool = False
        self._queued_tasks: list[asyncio.TimerHandle] = []

        # These will get filled in after initial authentication:
        self._client: ClientWebSocketResponse | None = None

    @property
    def connected(self) -> bool:
        """Return if currently connected to the websocket."""
        return self._client is not None and not self._client.closed

    @staticmethod
    def _add_callback(
        callback_list: list, callback: Callable[..., Any]
    ) -> Callable[..., None]:
        """Add a callback callback to a particular list."""
        callback_list.append(callback)

        def remove() -> None:
            """Remove the callback."""
            callback_list.remove(callback)

        return remove

    async def _async_receive_json(self) -> dict[str, Any]:
        """Receive a JSON response from the websocket server."""
        assert self._client
        msg = await self._client.receive(300)
        if msg.type in (WSMsgType.CLOSE, WSMsgType.CLOSED, WSMsgType.CLOSING):
            LOG.error(f"Websocket: Received event to close connection: {msg.type}")
            raise ConnectionClosedError("Connection was closed.")

        if msg.type == WSMsgType.ERROR:
            LOG.error(f"Websocket: Received error event, Connection failed: {msg.type}")
            raise ConnectionFailedError

        if msg.type != WSMsgType.TEXT:
            LOG.error(f"Websocket: Received invalid message: {msg}")
            raise InvalidMessageError(f"Received non-text message: {msg.type}")

        try:
            data = json.loads(msg.data[:-1])
        except ValueError as err:
            raise InvalidMessageError("Received invalid JSON") from err

        self._watchdog.trigger()

        return cast(Dict[str, Any], data)

    async def _async_send_json(self, payload: dict[str, Any]) -> None:
        """Send a JSON message to the websocket server.
        Raises NotConnectedError if client is not connected.
        """
        if not self.connected:
            raise NotConnectedError

        assert self._client

        if self._api.log_traces:
            LOG.debug(f"RAW Sending data to websocket server: {json.dumps(payload)}")
        # Hilo added a control character (chr(30)) at the end of each payload they send.
        # They also expect this char to be there at the end of every payload we send them.
        await self._client.send_str(json.dumps(payload) + chr(30))

    def _parse_message(self, msg: dict[str, Any]) -> None:
        """Parse an incoming message."""
        if self._api.log_traces:
            LOG.debug(f"RAW Received message from websocket: {msg}")
        if msg.get("type") == SignalRMsgType.PING:
            schedule_callback(self._async_pong)
            return
        if isinstance(msg, dict) and not len(msg):
            self._ready = True
            self._ready_event.set()
            LOG.info("Websocket: Ready for data")
            return
        event = websocket_event_from_payload(msg)
        for callback in self._event_callbacks:
            schedule_callback(callback, event)

    def add_connect_callback(self, callback: Callable[..., Any]) -> Callable[..., None]:
        """Add a callback callback to be called after connecting.
        :param callback: The method to call after connecting
        :type callback: ``Callable[..., None]``
        """
        return self._add_callback(self._connect_callbacks, callback)

    def add_disconnect_callback(
        self, callback: Callable[..., Any]
    ) -> Callable[..., None]:
        """Add a callback callback to be called after disconnecting.
        :param callback: The method to call after disconnecting
        :type callback: ``Callable[..., None]``
        """
        return self._add_callback(self._disconnect_callbacks, callback)

    def add_event_callback(self, callback: Callable[..., Any]) -> Callable[..., None]:
        """Add a callback callback to be called upon receiving an event.
        Note that callbacks should expect to receive a WebsocketEvent object as a
        parameter.
        :param callback: The method to call after receiving an event.
        :type callback: ``Callable[..., None]``
        """
        return self._add_callback(self._event_callbacks, callback)

    async def async_connect(self) -> None:
        """Connect to the websocket server."""
        if self.connected:
            LOG.debug("Websocket: async_connect() called but already connected")
            return

        LOG.info("Websocket: Connecting to server")
        if self._api.log_traces:
            LOG.debug(f"RAW Websocket URL: {self._api.full_ws_url}")
        headers = {
            "Sec-WebSocket-Extensions": "permessage-deflate; client_max_window_bits",
            "Pragma": "no-cache",
            "Cache-Control": "no-cache",
            "User-Agent": DEFAULT_USER_AGENT,
            "Origin": "http://localhost",
            "Accept-Language": "en-US,en;q=0.9",
        }
        # NOTE(dvd): for troubleshooting purpose we can pass a mitmproxy as env variable
        proxy_env: dict[str, Any] = {}
        if proxy := environ.get("WS_PROXY"):
            proxy_env["proxy"] = proxy
            proxy_env["verify_ssl"] = False

        try:
            self._client = await self._api.session.ws_connect(
                URL(
                    self._api.full_ws_url.replace("/DeviceHub", "%2FDeviceHub"),
                    encoded=True,
                ),
                heartbeat=55,
                headers=headers,
                **proxy_env,
            )
        except (ClientError, ServerDisconnectedError, WSServerHandshakeError) as err:
            LOG.error(f"Unable to connect to WS server {err}")
            if hasattr(err, "status") and err.status in (401, 403, 404, 409):  # type: ignore
                raise InvalidCredentialsError("Invalid credentials") from err
        except Exception as err:
            LOG.error(f"Unable to connect to WS server {err}")
            raise CannotConnectError(err) from err

        LOG.info("Connected to websocket server")
        self._watchdog.trigger()
        for callback in self._connect_callbacks:
            schedule_callback(callback)

    async def _clean_queue(self) -> None:
        for task in self._queued_tasks:
            task.cancel()

    async def async_disconnect(self) -> None:
        """Disconnect from the websocket server."""
        self._clean_queue()
        if not self.connected:
            return

        assert self._client

        await self._client.close()

        LOG.info("Disconnected from websocket server")

    async def async_listen(self) -> None:
        """Start listening to the websocket server."""
        assert self._client
        LOG.info("Websocket: Listen started.")
        try:
            while not self._client.closed:
                message = await self._async_receive_json()
                self._parse_message(message)
        except ConnectionClosedError as err:
            LOG.error(f"Websocket: Closed while listening: {err}")
            LOG.exception(err)
            pass
        finally:
            LOG.info("Websocket: Listen completed; cleaning up")
            self._watchdog.cancel()
            self._clean_queue()

            for callback in self._disconnect_callbacks:
                schedule_callback(callback)

    async def async_reconnect(self) -> None:
        """Reconnect (and re-listen, if appropriate) to the websocket."""
        LOG.warning("Websocket: Reconnecting")
        await self.async_disconnect()
        await asyncio.sleep(1)
        await self.async_connect()

    async def send_status(self) -> None:
        LOG.debug("Sending status")
        self._ready = False
        await self._async_send_json({"protocol": "json", "version": 1})

    async def _async_pong(self) -> None:
        await self._async_send_json({"type": SignalRMsgType.PING})

    async def async_invoke(
        self, arg: list, target: str, inv_id: int, inv_type: WSMsgType = WSMsgType.TEXT
    ) -> None:
        if not self._ready:
            LOG.warning(
                f"Delaying invoke {target} {inv_id} {arg}: Websocket not ready."
            )
            try:
                await asyncio.wait_for(self._ready_event.wait(), timeout=10)
            except asyncio.TimeoutError:
                return
            self._ready_event.clear()
        await self._async_send_json(
            {
                "arguments": arg,
                "invocationId": str(inv_id),
                "target": target,
                "type": inv_type,
            }
        )



from datetime import datetime
from typing import Any, Union

from pyhilo import API
from pyhilo.const import HILO_DEVICE_TYPES, LOG
from pyhilo.device import DeviceReading, HiloDevice
from pyhilo.device.climate import Climate  # noqa
from pyhilo.device.light import Light  # noqa
from pyhilo.device.sensor import Sensor  # noqa


class Devices:
    def __init__(self, api: API):
        self._api = api
        self.devices: list[HiloDevice] = []
        self.location_id: int = 0

    @property
    def all(self) -> list[HiloDevice]:
        return self.devices

    @property
    def attributes_list(self) -> list[Union[int, dict[int, list[str]]]]:
        """This is sent to websocket to subscribe to the device attributes updates

        :return: Dict of devices (key) with their attributes.
        :rtype: list
        """
        return [
            self.location_id,
            {
                d.id: d.hilo_attributes
                for d in self.devices
                if d.id > 1 and len(d.hilo_attributes)
            },
        ]

    def parse_values_received(self, values: list[dict[str, Any]]) -> list[HiloDevice]:
        readings = []
        for val in values:
            val["device_attribute"] = self._api.dev_atts(
                val.pop("attribute"), val.pop("valueType")
            )
            readings.append(DeviceReading(**val))
        return self._map_readings_to_devices(readings)

    def _map_readings_to_devices(
        self, readings: list[DeviceReading]
    ) -> list[HiloDevice]:
        updated_devices = []
        for reading in readings:
            if device := self.find_device(reading.device_id):
                device.readings = [r for r in device.readings if r != reading] + [
                    reading
                ]
                device.last_update = datetime.now()
                LOG.debug(f"{device} Received {reading}")
                if device not in updated_devices:
                    updated_devices.append(device)
            else:
                LOG.warning(
                    f"Unable to find device {reading.device_id} for reading {reading}"
                )
        return updated_devices

    def find_device(self, id: int) -> HiloDevice:
        return next((d for d in self.devices if d.id == id), None)  # type: ignore

    async def update(self) -> None:
        for device in await self._api.get_devices(self.location_id):
            device["location_id"] = self.location_id
            try:
                device_type = HILO_DEVICE_TYPES[device["type"]]
            except KeyError:
                LOG.warning(f"Unknown device type {device['type']}, adding as Sensor")
                device_type = "Sensor"
            klass = globals()[device_type]
            dev = self.find_device(device.get("id", 0)) or klass(self._api, **device)
            dev.update(**device)
            if dev not in self.devices:
                self.devices.append(dev)

    async def async_init(self) -> None:
        """Initialize the Hilo "manager" class."""
        LOG.info("Initialising after websocket is connected")
        self.location_id = await self._api.get_location_id()
        await self.update()



"""Define utility modules."""
import asyncio
from datetime import datetime, timedelta
import re
from typing import Any, Callable

from dateutil import tz
from dateutil.parser import parse

from pyhilo.const import LOG  # noqa: F401

CAMEL_REX_1 = re.compile("(.)([A-Z][a-z]+)")
CAMEL_REX_2 = re.compile("([a-z0-9])([A-Z])")


def schedule_callback(callback: Callable[..., Any], *args: Any) -> None:
    """Schedule a callback to be called."""
    if asyncio.iscoroutinefunction(callback):
        asyncio.create_task(callback(*args))
    else:
        loop = asyncio.get_running_loop()
        loop.call_soon(callback, *args)


def camel_to_snake(string: str) -> str:
    string = CAMEL_REX_1.sub(r"\1_\2", string)
    return CAMEL_REX_2.sub(r"\1_\2", string).lower()


def snake_to_camel(string: str) -> str:
    components = string.split("_")
    return components[0].title() + "".join(x.title() for x in components[1:])


def from_utc_timestamp(date_string: str) -> datetime:
    from_zone = tz.tzutc()
    to_zone = tz.tzlocal()
    return parse(date_string).replace(tzinfo=from_zone).astimezone(to_zone)


def time_diff(ts1: datetime, ts2: datetime) -> timedelta:
    to_zone = tz.tzlocal()
    return ts1.astimezone(to_zone) - ts2.astimezone(to_zone)



"""Climate object """

from typing import Union, cast

from pyhilo import API
from pyhilo.const import LOG
from pyhilo.device import HiloDevice


class Climate(HiloDevice):
    def __init__(self, api: API, **kwargs: dict[str, Union[str, int]]):
        super().__init__(api, **kwargs)
        LOG.debug(f"Setting up Climate device: {self.name}")

    @property
    def current_temperature(self) -> float:
        return cast(float, self.get_value("current_temperature", 0))

    @property
    def target_temperature(self) -> float:
        return cast(float, self.get_value("target_temperature", 0))

    @property
    def max_temp(self) -> float:
        return cast(float, self.get_value("max_temp_setpoint", 0))

    @property
    def min_temp(self) -> float:
        return cast(float, self.get_value("min_temp_setpoint", 0))

    @property
    def hvac_mode(self) -> str:
        attr = self.get_value("heating", 0)
        return "heat" if attr > 0 else "off"

    async def async_set_temperature(self, **kwargs: dict[str, int]) -> None:
        temperature = kwargs.get("temperature", 0)
        if temperature:
            LOG.info(f"{self._tag} Setting temperature to {temperature}")
            await self.set_attribute("target_temperature", temperature)  # type: ignore



"""Define devices"""
from __future__ import annotations

from dataclasses import dataclass, field
from datetime import datetime
from typing import TYPE_CHECKING, Any, Union, cast

from pyhilo.const import (
    HILO_DEVICE_ATTRIBUTES,
    HILO_LIST_ATTRIBUTES,
    HILO_PROVIDERS,
    HILO_READING_TYPES,
    HILO_UNIT_CONVERSION,
    LOG,
)
from pyhilo.util import camel_to_snake, from_utc_timestamp

if TYPE_CHECKING:
    from pyhilo import API


def get_device_attributes() -> list[DeviceAttribute]:
    attributes = []
    for attribute, value_type in HILO_READING_TYPES.items():
        attributes.append(DeviceAttribute(attribute, value_type))
    return attributes


class HiloDevice:
    def __init__(self, api: API, **kwargs: dict[str, Union[str, int]]) -> None:
        self._api = api
        self.id = 0
        self.location_id = 0
        self.type = "Unknown"
        self.name = "Unknown"
        self.model: Union[str, None] = None
        self.supported_attributes: list[DeviceAttribute] = []
        self.settable_attributes: list[DeviceAttribute] = []
        self.readings: list[DeviceReading] = []
        self.update(**kwargs)

    def update(self, **kwargs: dict[str, Union[str, int]]) -> None:
        # TODO(dvd): This has to be re-written, this is not dynamic at all.
        for orig_att, val in kwargs.items():
            att = camel_to_snake(orig_att)
            if att not in HILO_DEVICE_ATTRIBUTES:
                LOG.warning(f"Unknown device attribute {att}: {val}")
                continue
            elif att in HILO_LIST_ATTRIBUTES:
                # This is where we generated the supported_attributes and settable_attributes
                # list using the DeviceAttribute object.
                new_val: list[DeviceAttribute] = [
                    DeviceAttribute(k, HILO_READING_TYPES.get(k, ""))
                    for k in map(str.strip, val.split(","))  # type: ignore
                    if k and k != "None"
                ]
                if len(new_val) == 0:
                    # Some sensors like the OneLink FirstAlert don't have any attributes
                    # but they have a "Disconnected" attribute even though it doesn't show
                    # up in supported_attributes.
                    new_val.append(DeviceAttribute("Disconnected", "null"))
            elif att == "provider":
                att = "manufacturer"
                new_val = HILO_PROVIDERS.get(int(val), f"Unknown ({val})")  # type: ignore
            else:
                if att == "serial":
                    att = "identifier"
                elif att == "model_number":
                    att = "model"
                new_val = val  # type: ignore
            setattr(self, att, new_val)
        if self.type == "Thermostat" and not self.model:
            self.model = "EQ000016"
        if self.model in ["43082", "43078", "46199"]:
            self.manufacturer = "Jasco Enbrighten"
        self._tag = f"[{self.type} {self.name} ({self.id})]"
        self.last_update = datetime.now()

    async def set_attribute(self, attribute: str, value: Union[str, int, None]) -> None:
        if dev_attribute := self._api.dev_atts(attribute):
            LOG.debug(f"{self._tag} Setting {dev_attribute} to {value}")
            await self._set_attribute(dev_attribute, value)
            return
        LOG.warning(
            f"{self._tag} Unable to set attribute {attribute}: Unknown attribute"
        )

    async def _set_attribute(
        self, attribute: DeviceAttribute, value: Union[str, int, None]
    ) -> None:
        if attribute in self.settable_attributes:
            await self._api._set_device_attribute(self, attribute, value)
        else:
            LOG.warning(f"{self._tag} Invalid attribute {attribute} for device")

    def get_attribute(self, attribute: str) -> Union[DeviceReading, None]:
        if dev_attribute := self._api.dev_atts(attribute):
            return self._get_attribute(dev_attribute)
        LOG.warning(
            f"{self._tag} Unable to get attribute {attribute}: Unknown attribute"
        )
        return None

    def _get_attribute(self, attribute: DeviceAttribute) -> Union[DeviceReading, None]:
        reading = next(
            (r for r in self.readings if r.device_attribute == attribute), None
        )
        return reading

    def has_attribute(self, attr: str) -> bool:
        return next((True for k in self.supported_attributes if k.attr == attr), False)

    def get_value(
        self, attribute: str, default: Union[str, int, float, None] = None
    ) -> Any:
        attr = self.get_attribute(attribute)
        return attr.value if attr else default

    @property
    def hilo_attributes(self) -> list[str]:
        return [
            k.hilo_attribute
            for k in self.supported_attributes
            if k.hilo_attribute != "Humidity"
        ]

    @property
    def is_on(self) -> bool:
        return cast(bool, self.get_value("is_on"))

    @property
    def available(self) -> bool:
        return not self.get_value("disconnected") or False

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, HiloDevice):
            return NotImplemented
        return self.id == other.id

    def __str__(self) -> str:
        return self._tag


@dataclass(frozen=True, eq=True)
class DeviceAttribute:
    """Define a representation of an Attribute returned by Hilo."""

    hilo_attribute: str = field(compare=False)
    hilo_value_type: str = field(compare=False)
    attr: str | None = field(init=False, compare=True)
    value_type: str | None = field(init=False, compare=False)

    def __post_init__(self) -> None:
        if self.hilo_attribute == "OnOff":
            attr = "is_on"
        else:
            attr = camel_to_snake(self.hilo_attribute)
        if self.hilo_value_type in ("null", "OnOff"):
            value = "boolean"
        else:
            value = HILO_UNIT_CONVERSION.get(
                self.hilo_value_type, camel_to_snake(self.hilo_value_type)
            )
        object.__setattr__(self, "attr", attr)
        object.__setattr__(self, "value_type", value)


class DeviceReading:
    def __init__(self, **kwargs: dict[str, Any]):
        kwargs["timeStamp"] = from_utc_timestamp(kwargs.pop("timeStampUTC", ""))  # type: ignore
        self.id = 0
        self.value: Union[int, bool, str] = 0
        self.device_id = 0
        self.device_attribute: DeviceAttribute
        self.__dict__.update({camel_to_snake(k): v for k, v in kwargs.items()})
        self.unit_of_measurement = (
            self.device_attribute.value_type
            if self.device_attribute and self.device_attribute.value_type != "boolean"
            else ""
        )
        if not self.device_attribute:
            LOG.warning(f"Received invalid reading for {self.device_id}: {kwargs}")

    def __repr__(self) -> str:
        return f"<Reading {self.device_attribute.attr} {self.value}{self.unit_of_measurement}>"

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, DeviceReading):
            return NotImplemented
        return self.device_attribute.attr == other.device_attribute.attr
